#define _LARGEFILE64_SOURCE

#include <cstdio>
#include <cstdlib>
#include <string>
#include <vector>

#include <sys/stat.h>
#include <unistd.h>

#include <argh.hh>

using namespace std;

#define STAT stat64
#define OFFT __off64_t

class FileSpec
{
    string fn;
    OFFT size;
    bool ok;
public:
    FileSpec(const string &s) : fn(s)
    {
    }
    bool isok() const { return ok; }
    void GetSize()
    {
        struct STAT st;
        ok = !STAT(fn.c_str(), &st);
        if(ok)
        {
            if(!S_ISREG(st.st_mode))
            {
                fprintf(stderr,
                    "'%s' is troublesome.\n"
                    "This program does not have a solution for \"file is not a file\".\n"
                    "This of course is user error. Exit.\n",
                    fn.c_str());
                exit(-1);
            }
            size = st.st_size;
        }
        else
            perror(fn.c_str());
    }
    OFFT getsize() const { return size; }
    const string &getname() const { return fn; }
};

typedef vector<FileSpec> filelist;

class fileindex
{
    const filelist &files;
    vector<unsigned> list;
public:
    fileindex(const filelist &tab) : files(tab)
    {
    }
    
    void operator<< (unsigned n)
    {
        files.at(n); // verify bounds
        list.insert(list.begin(), n);
    }
    void operator>> (unsigned n)
    {
        files.at(n); // verify bounds
        list.push_back(n);
    }
    void operator-= (unsigned n)
    {
        for(unsigned a=0; a<list.size(); )
            if(list[a] == n)
                erase(a);
            else
                ++a;
    }
    unsigned size() const { return list.size(); }
    unsigned operator[] (unsigned n) const { return list.at(n); }
    void erase(unsigned n) { list.erase(list.begin() + n); }
    
    OFFT getsize() const
    {
        OFFT size = 0;
        for(unsigned a=0; a<list.size(); ++a)
            size += files[list[a]].getsize();
        return size;
    }
};

int main(int argc, const char *const *argv)
{
    ParamHandler Argh;
    
    OFFT cdsize = 650 << 20;
    
    filelist files;
    
    Argh.AddLong("help",    'h').SetBool().SetDesc("This help.");
    Argh.AddLong("version", 'V').SetBool().SetDesc("Displays version information.");

    Argh.StartParse(argc, argv);
    for(;;)
    {
        int c = Argh.GetParam();
        if(c == -1)break;
        switch(c)
        {
            case 'V': printf("%s\n", VERSION); return 0;
            case 'h':
                printf(
                    "This is software\n"
                    "\nUsage: software [<option> [<...>]] <file> [<...>]\n"
                    "\nThis software does something for the files.\n"
                    "\nOptions:\n");
                Argh.ListOptions();
                printf("\nNo warranty whatsoever.\n");
                fprintf(stderr, "This program does not have proper help texts written.\n");
                fprintf(stderr, "This is not user error. Exit.\n");
                return 0;
            default:
                files.push_back(Argh.GetString());
        }
    }
    if(!Argh.ok())return -1;
    
    if(!files.size())
    {
        fprintf(stderr, "This program does not have a solution for \"no files specified\" case.\n");
        fprintf(stderr, "This of course is user error. Exit.\n");
        return -1;
    }
    
    for(unsigned a=0; a<files.size(); ++a)
        files[a].GetSize();

    for(unsigned a=0; a<files.size(); ++a)
        if(!files[a].isok())
        {
            fprintf(stderr, "This program does not have a solution for \"input file sizes unknown\" case.\n");
            fprintf(stderr, "This of course is user error. Exit.\n");
            return -1;
        }
    
    OFFT total = 0;
    for(unsigned a=0; a<files.size(); ++a)
    {
        if(files[a].getsize() > cdsize)
        {
            fprintf(stderr,
                "'%s' is troublesome.\n"
                "This program does not have a solution for \"file bigger than disc\".\n"
                "This of course is user error. Exit.",
                    files[a].getname().c_str());
            return -1;
        }
        total += files[a].getsize();
    }
    
    unsigned cdcount = (total+cdsize-1) / cdsize;
    OFFT loss = cdcount * cdsize - total;
    
    printf("Number of CDs required at minimum: %u ", cdcount);
    printf(" (waste space total: %.1f MB)\n", loss / 1048576.0);
    
    vector<fileindex> cd(cdcount, fileindex(files));
    for(unsigned a=0, c=cdcount; a < files.size(); )
        for(--c; a < files.size(); ++a)
        {
            unsigned f = files.size() - a - 1;
            if(c)
            {
                OFFT size = cd[c].getsize();
                if(size + files[f].getsize() > cdsize)break;
            }
            cd[c] << f;
        }
    
    for(unsigned c=0; c<cdcount; ++c)
    {
        printf("CD %u: %.1f MB", c, cd[c].getsize() / 1048576.0);
        printf(" (files");
        for(unsigned a=0; a<cd[c].size(); ++a)
            printf(" %u", cd[c][a]);
        printf(")");
        printf("\n");
    }
    for(unsigned c=0; c<cdcount; ++c)
    {
    	if(CD[c].getsize() > cdsize)
    	{
    		printf("CD %u too big.\n", c);
    	}
    }
    
    fprintf(stderr, "Not implemented\n");
    return 0;
}
